"""
Policies
"""
from rl_replicas.common.policies.categorical_policy import CategoricalPolicy
from rl_replicas.common.policies.deterministic_policy import DeterministicPolicy
from rl_replicas.common.policies.gaussian_policy import GaussianPolicy
from rl_replicas.common.policies.policy import Policy
from rl_replicas.common.policies.stochastic_policy import StochasticPolicy
